<?php

namespace Drupal\blazy\Utility;

use Drupal\blazy\Blazy;

/**
 * Provides url, route, request, stream, or any path-related methods.
 */
class Path {

  /**
   * The AMP page.
   *
   * @var bool
   */
  private static $isAmp;

  /**
   * The preview mode to disable Blazy where JS is not available, or useless.
   *
   * @var bool
   */
  private static $isPreview;

  /**
   * The preview mode to disable interactive elements.
   *
   * @var bool
   */
  private static $isSandboxed;

  /**
   * Retrieves the file url generator service.
   *
   * @return \Drupal\Core\File\FileUrlGenerator
   *   The file url generator.
   *
   * @see https://www.drupal.org/node/2940031
   */
  public static function fileUrlGenerator() {
    return Blazy::service('file_url_generator');
  }

  /**
   * Retrieves the path resolver.
   *
   * @return \Drupal\Core\Extension\ExtensionPathResolver
   *   The path resolver.
   */
  public static function pathResolver() {
    return Blazy::service('extension.path.resolver');
  }

  /**
   * Retrieves the request stack.
   *
   * @return \Symfony\Component\HttpFoundation\RequestStack
   *   The request stack.
   */
  public static function requestStack() {
    return Blazy::service('request_stack');
  }

  /**
   * Retrieves the currently active route match object.
   *
   * @return \Drupal\Core\Routing\RouteMatchInterface
   *   The currently active route match object.
   */
  public static function routeMatch() {
    return Blazy::service('current_route_match');
  }

  /**
   * Retrieves the stream wrapper manager service.
   *
   * @return \Drupal\Core\StreamWrapper\StreamWrapperManager
   *   The stream wrapper manager.
   */
  public static function streamWrapperManager() {
    return Blazy::service('stream_wrapper_manager');
  }

  /**
   * Retrieves the request.
   *
   * @return \Symfony\Component\HttpFoundation\Request
   *   The request.
   *
   * @see https://github.com/symfony/symfony/blob/6.0/src/Symfony/Component/HttpFoundation/Request.php
   */
  public static function request() {
    if ($stack = self::requestStack()) {
      return $stack->getCurrentRequest();
    }
    return NULL;
  }

  /**
   * Returns the commonly used path, or just the base path.
   *
   * @todo remove drupal_get_path check when min D9.3.
   */
  public static function getPath($type, $name, $absolute = FALSE): ?string {
    if ($resolver = self::pathResolver()) {
      $path = $resolver->getPath($type, $name);
    }
    else {
      $function = 'drupal_get_path';
      $path = is_callable($function) ? $function($type, $name) : '';
    }
    return $absolute ? \base_path() . $path : $path;
  }

  /**
   * Returns multiple libraries keyed by its name.
   */
  public static function getLibraries(array $names, $base_path = FALSE): array {
    $libraries = [];
    foreach (self::libraries($names, TRUE) as $key => $path) {
      if ($path) {
        $libraries[$key] = $base_path ? \base_path() . $path : $path;
      }
    }
    return $libraries;
  }

  /**
   * Returns the first found library path.
   */
  public static function getLibrariesPath($name, $base_path = FALSE): ?string {
    $library = '';
    $names = is_array($name) ? $name : [$name];
    foreach (self::libraries($names) as $path) {
      if ($path) {
        $library = $base_path ? \base_path() . $path : $path;
        break;
      }
    }
    return $library;
  }

  /**
   * Checks if Blazy is in CKEditor preview mode where no JS assets are loaded.
   */
  public static function isPreview(): bool {
    if (!isset(self::$isPreview)) {
      self::$isPreview = self::isAmp() || self::isSandboxed();
    }
    return self::$isPreview;
  }

  /**
   * Checks if Blazy is in AMP pages.
   */
  public static function isAmp(): bool {
    if (!isset(self::$isAmp)) {
      $request = self::request();
      self::$isAmp = $request && $request->query->get('amp');
    }
    return self::$isAmp;
  }

  /**
   * In CKEditor without JS assets, interactive elements must be sandboxed.
   */
  public static function isSandboxed(): bool {
    if (!isset(self::$isSandboxed)) {
      $check = FALSE;
      if ($router = self::routeMatch()) {
        if ($route = $router->getRouteName()) {
          $edits = ['entity_browser.', 'edit_form', 'add_form', '.preview'];
          foreach ($edits as $key) {
            if (Blazy::has($route, $key)) {
              $check = TRUE;
              break;
            }
          }
        }
      }

      self::$isSandboxed = $check;
    }
    return self::$isSandboxed;
  }

  /**
   * Provides a wrapper to replace deprecated libraries_get_path() at ease.
   */
  private static function libraries(array $libraries, $keyed = FALSE): \Generator {
    if ($finder = Blazy::service('library.libraries_directory_file_finder')) {
      foreach ($libraries as $library) {
        $result = $finder->find($library);
        if ($keyed) {
          yield $library => $result;
        }
        else {
          yield $result;
        }
      }
    }
    else {
      // @todo remove when min D9.2.
      $dep = 'libraries_get_path';
      foreach ($libraries as $library) {
        $result = is_callable($dep) ? $dep($library) : '';
        if ($keyed) {
          yield $library => $result;
        }
        else {
          yield $result;
        }
      }
    }
  }

}
